@@ -6,15 +6,19 @@ import (
 	"net/http"
 	"os"
 	"strings"
+	"time"
 
 	"github.com/sagernet/sing-box/adapter"
 	"github.com/sagernet/sing-box/common/tls"
 	C "github.com/sagernet/sing-box/constant"
 	"github.com/sagernet/sing-box/option"
 	"github.com/sagernet/sing/common"
+	"github.com/sagernet/sing/common/buf"
+	"github.com/sagernet/sing/common/bufio"
 	E "github.com/sagernet/sing/common/exceptions"
 	M "github.com/sagernet/sing/common/metadata"
 	N "github.com/sagernet/sing/common/network"
+	aTLS "github.com/sagernet/sing/common/tls"
 	sHttp "github.com/sagernet/sing/protocol/http"
 
 	"golang.org/x/net/http2"
@@ -24,32 +28,34 @@ import (
 var _ adapter.V2RayServerTransport = (*Server)(nil)
 
 type Server struct {
-	ctx          context.Context
-	handler      N.TCPConnectionHandler
-	errorHandler E.Handler
-	httpServer   *http.Server
-	h2Server     *http2.Server
-	h2cHandler   http.Handler
-	host         []string
-	path         string
-	method       string
-	headers      http.Header
+	ctx        context.Context
+	tlsConfig  tls.ServerConfig
+	handler    adapter.V2RayServerTransportHandler
+	httpServer *http.Server
+	h2Server   *http2.Server
+	h2cHandler http.Handler
+	host       []string
+	path       string
+	method     string
+	headers    http.Header
 }
 
 func (s *Server) Network() []string {
 	return []string{N.NetworkTCP}
 }
 
-func NewServer(ctx context.Context, options option.V2RayHTTPOptions, tlsConfig tls.ServerConfig, handler N.TCPConnectionHandler, errorHandler E.Handler) (*Server, error) {
+func NewServer(ctx context.Context, options option.V2RayHTTPOptions, tlsConfig tls.ServerConfig, handler adapter.V2RayServerTransportHandler) (*Server, error) {
 	server := &Server{
-		ctx:          ctx,
-		handler:      handler,
-		errorHandler: errorHandler,
-		h2Server:     new(http2.Server),
-		host:         options.Host,
-		path:         options.Path,
-		method:       options.Method,
-		headers:      make(http.Header),
+		ctx:       ctx,
+		tlsConfig: tlsConfig,
+		handler:   handler,
+		h2Server: &http2.Server{
+			IdleTimeout: time.Duration(options.IdleTimeout),
+		},
+		host:    options.Host,
+		path:    options.Path,
+		method:  options.Method,
+		headers: make(http.Header),
 	}
 	if server.method == "" {
 		server.method = "PUT"
@@ -58,7 +64,7 @@ func NewServer(ctx context.Context, options option.V2RayHTTPOptions, tlsConfig t
 		server.path = "/" + server.path
 	}
 	for key, value := range options.Headers {
-		server.headers.Set(key, value)
+		server.headers[key] = value
 	}
 	server.httpServer = &http.Server{
 		Handler:           server,
@@ -66,13 +72,6 @@ func NewServer(ctx context.Context, options option.V2RayHTTPOptions, tlsConfig t
 		MaxHeaderBytes:    http.DefaultMaxHeaderBytes,
 	}
 	server.h2cHandler = h2c.NewHandler(server, server.h2Server)
-	if tlsConfig != nil {
-		stdConfig, err := tlsConfig.Config()
-		if err != nil {
-			return nil, err
-		}
-		server.httpServer.TLSConfig = stdConfig
-	}
 	return server, nil
 }
 
@@ -83,18 +82,15 @@ func (s *Server) ServeHTTP(writer http.ResponseWriter, request *http.Request) {
 	}
 	host := request.Host
 	if len(s.host) > 0 && !common.Contains(s.host, host) {
-		writer.WriteHeader(http.StatusBadRequest)
-		s.badRequest(request, E.New("bad host: ", host))
+		s.fallbackRequest(request.Context(), writer, request, http.StatusBadRequest, E.New("bad host: ", host))
 		return
 	}
 	if !strings.HasPrefix(request.URL.Path, s.path) {
-		writer.WriteHeader(http.StatusNotFound)
-		s.badRequest(request, E.New("bad path: ", request.URL.Path))
+		s.fallbackRequest(request.Context(), writer, request, http.StatusNotFound, E.New("bad path: ", request.URL.Path))
 		return
 	}
 	if request.Method != s.method {
-		writer.WriteHeader(http.StatusNotFound)
-		s.badRequest(request, E.New("bad method: ", request.Method))
+		s.fallbackRequest(request.Context(), writer, request, http.StatusNotFound, E.New("bad method: ", request.Method))
 		return
 	}
 
@@ -112,16 +108,24 @@ func (s *Server) ServeHTTP(writer http.ResponseWriter, request *http.Request) {
 	var metadata M.Metadata
 	metadata.Source = sHttp.SourceAddress(request)
 	if h, ok := writer.(http.Hijacker); ok {
-		conn, _, err := h.Hijack()
+		conn, reader, err := h.Hijack()
 		if err != nil {
-			writer.WriteHeader(http.StatusInternalServerError)
-			s.badRequest(request, E.Cause(err, "hijack conn"))
+			s.fallbackRequest(request.Context(), writer, request, http.StatusInternalServerError, E.Cause(err, "hijack conn"))
 			return
 		}
+		if cacheLen := reader.Reader.Buffered(); cacheLen > 0 {
+			cache := buf.NewSize(cacheLen)
+			_, err = cache.ReadFullFrom(reader.Reader, cacheLen)
+			if err != nil {
+				s.fallbackRequest(request.Context(), writer, request, http.StatusInternalServerError, E.Cause(err, "read cache"))
+				return
+			}
+			conn = bufio.NewCachedConn(conn, cache)
+		}
 		s.handler.NewConnection(request.Context(), conn, metadata)
 	} else {
 		conn := NewHTTP2Wrapper(&ServerHTTPConn{
-			newHTTPConn(request.Body, writer),
+			NewHTTPConn(request.Body, writer),
 			writer.(http.Flusher),
 		})
 		s.handler.NewConnection(request.Context(), conn, metadata)
@@ -129,24 +133,23 @@ func (s *Server) ServeHTTP(writer http.ResponseWriter, request *http.Request) {
 	}
 }
 
-func (s *Server) badRequest(request *http.Request, err error) {
-	s.errorHandler.NewError(request.Context(), E.Cause(err, "process connection from ", request.RemoteAddr))
+func (s *Server) fallbackRequest(ctx context.Context, writer http.ResponseWriter, request *http.Request, statusCode int, err error) {
+	conn := NewHTTPConn(request.Body, writer)
+	fErr := s.handler.FallbackConnection(ctx, &conn, M.Metadata{})
+	if fErr == nil {
+		return
+	} else if fErr == os.ErrInvalid {
+		fErr = nil
+	}
+	writer.WriteHeader(statusCode)
+	s.handler.NewError(request.Context(), E.Cause(E.Errors(err, E.Cause(fErr, "fallback connection")), "process connection from ", request.RemoteAddr))
 }
 
 func (s *Server) Serve(listener net.Listener) error {
-	fixTLSConfig := s.httpServer.TLSConfig == nil
-	err := http2.ConfigureServer(s.httpServer, s.h2Server)
-	if err != nil {
-		return err
-	}
-	if fixTLSConfig {
-		s.httpServer.TLSConfig = nil
-	}
-	if s.httpServer.TLSConfig == nil {
-		return s.httpServer.Serve(listener)
-	} else {
-		return s.httpServer.ServeTLS(listener, "", "")
+	if s.tlsConfig != nil {
+		listener = aTLS.NewListener(listener, s.tlsConfig)
 	}
+	return s.httpServer.Serve(listener)
 }
 
 func (s *Server) ServePacket(listener net.PacketConn) error {
